
#import "Friio.h"
#include "b_cas_card_error_code.h"
#include <signal.h>


@implementation Friio

// 設定値定義
// -----------------------------------------------------------
static const int PIPE_REF = 1;
static const int BUFFER_SIZE = 327680;
static const int RESPONSE_SIZE = 16;

static int RESPONSE_NUM = 0;
static int THREAD_FLAG = 0;
static int TIMER_FLAG = 0;

static int useb25 = 0; //1:use

// レスポンス配列を生成
static USBResponse res[16];

// MULTI2デコード処理
ARIB_STD_B25 *b25;
B_CAS_CARD   *bcas;


// オープンメソッド
// -----------------------------------------------------------
- (int)open:(int)type lid:(const char*)lid
{
	int result = 0;

	find_type=type;
	Lid=lid;

	// インターフェースを検索
	result = [self _findDeviceInterface];
	if(!result) {
		return 0;
	}
	switch(friio_type){
	case FRIIO_WHITE:
		printf("white friio is found.\n");
		break;
	case FRIIO_BLACK:
		printf("black friio is found.\n");
	}

	// ストリームの初期化
	[self initStream];

	// 初期化B
	[self initStageB];

	return 1;
}




// デバイスリファレンス取得メソッド
// -----------------------------------------------------------
- (int)_findDeviceInterface
{
	SInt32					idVendor = 0x7A69;
    SInt32					idProduct = 0x0001;
	kern_return_t			error;
	CFNumberRef				numberRef;
	CFMutableDictionaryRef 	matchingDictionary = 0;
    io_iterator_t			iterator = 0;
	int						rc = 1;
	
	// マスターポートを生成
	error = IOMasterPort(MACH_PORT_NULL, &masterPort);
	if(error) {
		printf("_findDevice: Could not open master port.\n");
		return 0;
	}
	
	// 辞書を生成
	matchingDictionary = IOServiceMatching(kIOUSBDeviceClassName);
	// ベンダーIDを辞書に登録
	numberRef = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &idVendor);
	CFDictionaryAddValue(matchingDictionary, CFSTR(kUSBVendorID), numberRef);
    CFRelease(numberRef);
	numberRef = 0;
	// プロダクトIDを辞書に登録
    numberRef = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &idProduct);
	CFDictionaryAddValue(matchingDictionary, CFSTR(kUSBProductID), numberRef);
    CFRelease(numberRef);
    numberRef = 0;
	
	// Friioを探す
	error = IOServiceGetMatchingServices(masterPort, matchingDictionary, &iterator);
	matchingDictionary = 0;
	if(error) {
		printf("_findDevice: Unable to find the device of the friio.\n");
		return 0;
	}
	
	friio_type=0;
	
	// USBデバイスのリファレンスを取得
	//usbDeviceRef = IOIteratorNext(iterator);
	// インターフェースリファレンスの取得
	while( usbDeviceRef = IOIteratorNext(iterator) ){
		SInt32						score;
		IOReturn					iferror;
		IOCFPlugInInterface			**iodev;
		IOUSBFindInterfaceRequest	interfaceRequest;
		io_iterator_t				ifiterator;
		char						szLocID[2*sizeof(UInt32)+1];
		// プラグインインターフェースの生成
		iferror = IOCreatePlugInInterfaceForService(
			usbDeviceRef,
			kIOUSBDeviceUserClientTypeID,
			kIOCFPlugInInterfaceID,
			&iodev,
			&score
		);
		if(iferror) {
			printf("_findInterface: Unable to create a device interface. %x\n", error);
			//失敗したら次を探す
			IOObjectRelease(usbDeviceRef);
			continue;
		}
		iferror = (*iodev)->QueryInterface(
			iodev, 
			CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID245),
			(LPVOID)&dev
		);
//		if(iodev) (*iodev)->Release(iodev);
//		iodev=0;
		IODestroyPlugInInterface(iodev);
		if(iferror || !dev) {
			printf("_findInterface: Unable to create a device interface.\n");
			//失敗したら次を探す
			IOObjectRelease(usbDeviceRef);
			continue;
		}
		UInt32 locationID = 0;
		
		(*dev)->GetLocationID (dev, &locationID);
		sprintf(szLocID,"%08x",(int)locationID);
//		printf ("LocationID=%08x ", (int) locationID);
	
		// デバイスのオープン
		iferror = (*dev)->USBDeviceOpen(dev);
		if(iferror) {
			IOObjectRelease(usbDeviceRef);
			// デバイスの解放
			(*dev)->Release(dev);
			//失敗したら次を探す
			continue;
		}
	
		// インターフェースイテレータの生成
		interfaceRequest.bInterfaceClass = kIOUSBFindInterfaceDontCare;
		interfaceRequest.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
		interfaceRequest.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
		interfaceRequest.bAlternateSetting = kIOUSBFindInterfaceDontCare;
		error = (*dev)->CreateInterfaceIterator(dev, &interfaceRequest, &ifiterator);
		(*dev)->USBDeviceClose(dev);
		(*dev)->Release(dev);
		if(error) {
			printf("_findInterface: Unable to create interface iterator.\n");
			//失敗したら次を探す
			IOObjectRelease(usbDeviceRef);
			continue;
		}
		

		usbInterfaceRef = IOIteratorNext(ifiterator);
		
		// 後処理
		IOObjectRelease(ifiterator);
		ifiterator = 0;

		// インターフェースをセットアップ
		int result = [self _setupInterface];
		if(!result) {
			//失敗したら次を探す
			//IOObjectRelease(usbInterfaceRef);
			IOObjectRelease(usbDeviceRef);
			continue;
		}

		//初期化A
		[self initStageA];
		//自動選択の場合は見つかったら使う。もしくは、指定したfriioのタイプだったら。
		if( find_type == FRIIO_ANY || friio_type == find_type ){
			if( Lid ) if( strcmp( szLocID, Lid ) ){
				IOObjectRelease(usbDeviceRef);
				[self close];
				continue;
			}
			break;
		}
		//違ったら次ぎいってみよー
		IOObjectRelease(usbDeviceRef);
		[self close];
		continue;
	}
	if(!usbDeviceRef) rc=0;
		
	// 後処理
	IOObjectRelease(usbDeviceRef);
	usbDeviceRef=0;
	
	IOObjectRelease(iterator);
	iterator = 0;

	mach_port_deallocate(mach_task_self(), masterPort);
	
	return rc;
}


// インターフェースセットアップメソッド
// -----------------------------------------------------------
- (int)_setupInterface
{
    SInt32					score;
	IOReturn				error;
    IOCFPlugInInterface 	**iodev;

	// プラグインインターフェースの生成
	error = IOCreatePlugInInterfaceForService(
		usbInterfaceRef,
		kIOUSBInterfaceUserClientTypeID,
		kIOCFPlugInInterfaceID,
		&iodev,
		&score
	);
	IOObjectRelease(usbInterfaceRef);
	usbInterfaceRef=0;
	error = (*iodev)->QueryInterface(
		iodev,
		CFUUIDGetUUIDBytes(kIOUSBInterfaceInterfaceID245),
		(LPVOID)&interface
	);
	if(iodev) (*iodev)->Release(iodev);
	//IODestroyPlugInInterface(iodev);
	if(error || !interface){
		printf("_setupInterface: Unable to create a device interface.\n");
		return 0;
    }
	
	// インターフェースを開く
	error = (*interface)->USBInterfaceOpen(interface);
	if(error) {
		if (error == kIOReturnExclusiveAccess) {
			//printf("_setupInterface: interface already in use. %x\n", error);
			//使用中のデバイスはそのまま放置。
		} else {
			//printf("_setupInterface: Unable to open interface. %x\n", error);
		}
		(*interface)->Release(interface);
		interface=0;
		//printf("_setupInterface: Unable to open interface.\n");
		return 0;
    }
	
	return 1;
}


// クローズメソッド
// -----------------------------------------------------------
- (int)close
{
    IOReturn error;
	
	if(interface){
		// インターフェースを閉じる
		error = (*interface)->USBInterfaceClose(interface);
		if(error) {
			printf("close: Unable to close interface.\n");
			return 0;
		}
		// インターフェースを解放
		error = (*interface)->Release(interface);
		if(error) {
			printf("close: Unable to release interface.\n");
			return 0;
		}
	}

	return 1;
}


// リクエスト送信関数
// -----------------------------------------------------------
- (int)_sendRequest:(UInt16)wValue withIndex:(UInt16)wIndex
{
	unsigned char	iobuff[32];
	IOReturn		error;
	
	// リクエストを生成
	IOUSBDevRequest req;
	req.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
	req.bRequest = 0x01;
	req.wValue = wValue;
	req.wIndex = wIndex;
	req.wLength = 0;
	req.wLenDone = 0;
	req.pData = iobuff;
	
	// Friioに送信
	error = (*interface)->ControlRequest(interface, 0, &req);
	if(error) {
		return 0;
	}
	
	return 1;
}	


// 初期化処理Aメソッド
// -----------------------------------------------------------
- (int)initStageA
{
	int			error_flag;
	IOReturn	error;

	// リクエスト送信
	error_flag = [self _sendRequest:0x0002 withIndex:0x0011];
	// リクエストの送信に失敗したら初期化失敗
	if(!error_flag) {
		return 0;
	}
	[self _sendRequest:0x0000 withIndex:0x0011];
	
	unsigned char iobuff1[32];
	IOUSBDevRequest req1;
	
	memset(iobuff1, 0, sizeof(iobuff1));
	iobuff1[0]=0xC1;
	req1.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
	req1.bRequest = 0x03;
	req1.wValue = 0x3000;
	req1.wIndex = 0x00FE;
	req1.wLength = 1;
	req1.wLenDone = 0;
	req1.pData = iobuff1;
	error = (*interface)->ControlRequest(interface, 0, &req1);

	iobuff1[0]=0;
	req1.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
	req1.bRequest = 0x02;
	req1.wValue = 0x3000;
	req1.wIndex = 0x0100;
	req1.wLength = 1;
	req1.wLenDone = 0;
	req1.pData = iobuff1;
	error = (*interface)->ControlRequest(interface, 0, &req1);

	//White Friio
	if(iobuff1[0]){
		friio_type=FRIIO_WHITE;
		//fprintf(stderr,"WHITE\n");
	}else{	
		friio_type=FRIIO_BLACK;
		//fprintf(stderr,"BLACK\n");
		unsigned char iobuff2[32];
		IOUSBDevRequest req2;
		req2.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
		req2.bRequest = 0x02;
		req2.wValue = 0x3000;
		req2.wIndex = 0x0080;
		req2.wLength = 1;
		req2.wLenDone = 0;
		req2.pData = iobuff2;
		error = (*interface)->ControlRequest(interface, 0, &req2);
	}
	
	[self _sendRequest:0x0004 withIndex:0x0030];
	
	return 1;
}


// 初期化処理Aメソッド
// -----------------------------------------------------------
- (int)initStageB
{
	if(friio_type==FRIIO_WHITE){
		//White Friio initstageB
		[self _sendRequest:0x3040 withIndex:0x01];
		[self _sendRequest:0x3038 withIndex:0x04];
		[self _sendRequest:0x3040 withIndex:0x05];
		[self _sendRequest:0x3040 withIndex:0x07];
		[self _sendRequest:0x304F withIndex:0x0F];
		[self _sendRequest:0x3021 withIndex:0x11];
		[self _sendRequest:0x300B withIndex:0x12];
		[self _sendRequest:0x302F withIndex:0x13];
		[self _sendRequest:0x3031 withIndex:0x14];
		[self _sendRequest:0x3002 withIndex:0x16];
		[self _sendRequest:0x30C4 withIndex:0x21];
		[self _sendRequest:0x3020 withIndex:0x22];
		[self _sendRequest:0x3079 withIndex:0x2C];
		[self _sendRequest:0x3034 withIndex:0x2D];
		[self _sendRequest:0x3000 withIndex:0x2F];
		[self _sendRequest:0x3028 withIndex:0x30];
		[self _sendRequest:0x3031 withIndex:0x31];
		[self _sendRequest:0x30DF withIndex:0x32];
		[self _sendRequest:0x3001 withIndex:0x38];
		[self _sendRequest:0x3078 withIndex:0x39];
		[self _sendRequest:0x3033 withIndex:0x3B];
		[self _sendRequest:0x3033 withIndex:0x3C];
		[self _sendRequest:0x3090 withIndex:0x48];
		[self _sendRequest:0x3068 withIndex:0x51];
		[self _sendRequest:0x3038 withIndex:0x5E];
		[self _sendRequest:0x3000 withIndex:0x71];
		[self _sendRequest:0x3008 withIndex:0x72];
		[self _sendRequest:0x3000 withIndex:0x77];
		[self _sendRequest:0x3021 withIndex:0xC0];
		[self _sendRequest:0x3010 withIndex:0xC1];
		[self _sendRequest:0x301A withIndex:0xE4];
		[self _sendRequest:0x301F withIndex:0xEA];
		[self _sendRequest:0x3000 withIndex:0x77];
		[self _sendRequest:0x3000 withIndex:0x71];
		[self _sendRequest:0x3000 withIndex:0x71];
		[self _sendRequest:0x300C withIndex:0x76];
	}else{
		//Black friio initstageB
		[self _sendRequest:0x3040 withIndex:0x0001];
		[self _sendRequest:0x3030 withIndex:0x0004];
		[self _sendRequest:0x3040 withIndex:0x0005];
		[self _sendRequest:0x3040 withIndex:0x0007];
		[self _sendRequest:0x30ff withIndex:0x000f];
		[self _sendRequest:0x3027 withIndex:0x0011];
		[self _sendRequest:0x302f withIndex:0x0013];
		[self _sendRequest:0x3031 withIndex:0x0014];
		[self _sendRequest:0x3002 withIndex:0x0016];
		[self _sendRequest:0x3020 withIndex:0x0022];
		[self _sendRequest:0x3070 withIndex:0x0024];
		[self _sendRequest:0x3079 withIndex:0x002c];
		[self _sendRequest:0x3034 withIndex:0x002d];
		[self _sendRequest:0x3028 withIndex:0x0030];
		[self _sendRequest:0x3031 withIndex:0x0031];
		[self _sendRequest:0x3001 withIndex:0x0038];
		[self _sendRequest:0x3078 withIndex:0x0039];
		[self _sendRequest:0x3090 withIndex:0x0048];
		[self _sendRequest:0x3068 withIndex:0x0051];
		[self _sendRequest:0x3038 withIndex:0x005e];
		[self _sendRequest:0x3021 withIndex:0x00c0];
		[self _sendRequest:0x3010 withIndex:0x00c1];
		[self _sendRequest:0x301a withIndex:0x00e4];
		[self _sendRequest:0x301f withIndex:0x00ea];	
	}
	return 1;
}


// ストリーム初期化メソッド
// -----------------------------------------------------------
- (int)initStream
{
	[self _sendRequest:0x08 withIndex:0x33];
	[self _sendRequest:0x40 withIndex:0x37];
	[self _sendRequest:0x1F withIndex:0x3A];
	[self _sendRequest:0xFF withIndex:0x3B];
	[self _sendRequest:0x1F withIndex:0x3C];
	[self _sendRequest:0xFF withIndex:0x3D];
	[self _sendRequest:0x00 withIndex:0x38];
	[self _sendRequest:0x00 withIndex:0x35];
	[self _sendRequest:0x00 withIndex:0x39];
	[self _sendRequest:0x00 withIndex:0x36];
	return 1;
}

- (void)setb25:(int)ib25
{
	useb25 = ib25;
}

//signal対応
void stopRecording(){
	TIMER_FLAG=0;
}

// 録画実行メソッド
// -----------------------------------------------------------
- (int)rec:(char*)file withTime:(int)rec_sec
{
	// タイマーフラグとスレッドフラグを初期化
	TIMER_FLAG = 1;
	THREAD_FLAG = 1;

	bsd_signal(SIGINT,stopRecording);
	bsd_signal(SIGKILL,stopRecording);
	bsd_signal(SIGTERM,stopRecording);

	if(useb25){
		// b25インスタンス生成
		b25 = create_arib_std_b25();
		b25->set_multi2_round(b25, 4);
		b25->set_strip(b25, 0);
		b25->set_emm_proc(b25, 0);	
	
		// B-CASインスタンス生成
		bcas = create_b_cas_card();
		if(bcas->init(bcas)<0) useb25=false;
		// b25にB-CASを設定
		b25->set_b_cas_card(b25, bcas);
		if(!useb25) printf("No B-CAS Card\nNo b25 decode\n");
	}else{
		printf("No b25 decode\n");
	}
	
	printf("Recording\n");
	fflush(stdout);
	
	// ファイルパスを保存
	file_path = file;

#if defined(MAC_OS_X_VERSION_10_5) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
	// スレッドを生成
	NSThread *worker = [[NSThread alloc] initWithTarget:self selector:@selector(_recWorker:) object:nil];
	// スレッド起動
	[worker start];
#else
	// スレッドをデタッチ
	[NSThread detachNewThreadSelector:@selector(_recWorker:) toTarget:self withObject:nil];
#endif

	// 指定された時間待機
	time_t time_start = time(NULL);
	while(TIMER_FLAG && (time(NULL) <= time_start + rec_sec || rec_sec==0 )) {
		float siglvl;
//		printf(".");
		fflush(stdout);
		sleep(1);
		siglvl=[self getSignalLevel];
		printf("Signal level: %f    \r",siglvl);
	}
	printf("\n");
	
	// タイマーフラグを無効化
	TIMER_FLAG = 0;
	
	// スレッドを待つ
	while(THREAD_FLAG) {
		sleep(1);
	}

#if defined(MAC_OS_X_VERSION_10_5) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
	// スレッド解放
	[worker release];
#endif

	return 0;
}


// 録画読み込みスレッド用メソッド
// -----------------------------------------------------------
- (void)_recWorker:(id)obj
{
	int					i;
	FILE				*output_fd;
	IOReturn			error;
    CFRunLoopSourceRef	cfSource;

	// 非同期イベントソースの生成
	error = (*interface)->CreateInterfaceAsyncEventSource(interface, &cfSource);
	if(error) {
		printf("_recWorker: Unable to create event source.\n");
		return;
    }
	// イベントソースを登録
	CFRunLoopAddSource(CFRunLoopGetCurrent(), cfSource, kCFRunLoopDefaultMode);
	
	// 出力先のファイルデスクリプタを取得
	output_fd = fopen(file_path, "w");
	if(output_fd == NULL) {
		printf("_recWorker: Unable to open output file.\n");
		CFRunLoopRemoveSource(CFRunLoopGetCurrent(), cfSource, kCFRunLoopDefaultMode);
		THREAD_FLAG = 0;
		return;
	}
	
	// バッファの初期化
	for(i=0; i<RESPONSE_SIZE; i++) {
		// インターフェースを初期化
		res[i].interface = interface;
		// バッファを初期化
		memset(res[i].buffer, '\0', sizeof(res[i].buffer));
		// ファイルデスクリプタを保存
		res[i].fd = output_fd;
	}
	
	// おまじない
	(*interface)->SetAlternateInterface(interface, 1);
	(*interface)->SetAlternateInterface(interface, 0);

	// パイプ初期化
	error = (*interface)->ClearPipeStall(interface, (UInt8)1);
	if(error) {
		printf("_recWorker: Could not clear the pipe.\n");
	}
	
	// 非同期Bulk転送 1本目
	error = (*interface)->ReadPipeAsync(interface, 1, res[RESPONSE_NUM].buffer, BUFFER_SIZE, (IOAsyncCallback1)_workerCallback, &res[RESPONSE_NUM]);
	if(error) {
		printf("_recWorker: Read pipe error.\n");
		// 転送失敗時の後始末
		CFRunLoopRemoveSource(CFRunLoopGetCurrent(), cfSource, kCFRunLoopDefaultMode);
		fflush(output_fd);
		fclose(output_fd);
		THREAD_FLAG = 0;
		return;
	}
	
	// 非同期Bulk転送 2本目
	RESPONSE_NUM++;
	error = (*interface)->ReadPipeAsync(interface, 1, res[RESPONSE_NUM].buffer, BUFFER_SIZE, (IOAsyncCallback1)_workerCallback, &res[RESPONSE_NUM]);
	if(error) {
		printf("_recWorker: Read pipe error.\n");
		// 転送失敗時の後始末
		CFRunLoopRemoveSource(CFRunLoopGetCurrent(), cfSource, kCFRunLoopDefaultMode);
		fflush(output_fd);
		fclose(output_fd);
		THREAD_FLAG = 0;
		return;
	}
	
	// イベントループを実行
	CFRunLoopRun();

	// イベントループの後処理
	CFRunLoopRemoveSource(CFRunLoopGetCurrent(), cfSource, kCFRunLoopDefaultMode);
	fflush(output_fd);
	fclose(output_fd);
	THREAD_FLAG = 0;

}


// 録画読み込みスレッド用コールバック
// -----------------------------------------------------------
void _workerCallback(USBResponse *readRes, IOReturn result, void *arg) {
	IOReturn error;
	
	// コールバック結果のチェック
	if(result != kIOReturnSuccess) {
		printf("_workerCallback: Callback failed(%x).\n", result);
		CFRunLoopStop(CFRunLoopGetCurrent());
		return;
	}
	
	// パイプをクリア
	/*error = (*readRes->interface)->ClearPipeStall(readRes->interface, (UInt8)1);
	if(error) {
		printf("_workerCallback: Could not clear the pipe.\n");
	}*/
	
	// レスポンス番号を循環
	if(RESPONSE_NUM < RESPONSE_SIZE-1) {
		RESPONSE_NUM++;
	} else {
		RESPONSE_NUM = 0;
	}
	
	// 非同期Bulk転送を再実行
	error = (*res[RESPONSE_NUM].interface)->ReadPipeAsync(
		res[RESPONSE_NUM].interface,
		1,
		res[RESPONSE_NUM].buffer,
		BUFFER_SIZE,
		(IOAsyncCallback1)_workerCallback,
		&res[RESPONSE_NUM]
	);
	if(error) {
		printf("_workerCallback: Read pipe error.\n");
	}
	
	if(useb25){
	
		ARIB_STD_B25_BUFFER sbuf;
		ARIB_STD_B25_BUFFER dbuf;
	
		sbuf.data = readRes->buffer;
		sbuf.size = BUFFER_SIZE;
	
		b25->put(b25, &sbuf);
		b25->get(b25, &dbuf);
	
		// ファイルに書き込み
		fwrite(dbuf.data, 1, dbuf.size, readRes->fd);
	}else{
		fwrite(readRes->buffer, 1, BUFFER_SIZE, readRes->fd);
	}
	// 使ったバッファの初期化
	memset(readRes->buffer, '\0', sizeof(readRes->buffer));
	
	// 終了処理
	if(!TIMER_FLAG) {
		CFRunLoopStop(CFRunLoopGetCurrent());
	}
}

- (int)setChannel:(int)channel
{
	// 白の場合
	if( friio_type==FRIIO_WHITE ){ 
		IOReturn error;
		// チャンネルのチェック
		if(channel < 13 || channel > 62) {
			printf("setChannel: Invalid channel.\n");
			return 0;
		}
	
		// チャンネルデータ計算
		const uint32_t chCode = 0x0E7F + 0x002A * (uint32_t)(channel - 13);
	
		// リクエスト用バッファ
		uint8_t sbuf[2][5];
		memset(sbuf, '\0', sizeof(sbuf));
		uint8_t rbuf[2][1];
		memset(rbuf, '\0', sizeof(rbuf));
	
		// 1回目リクエスト設定	
		sbuf[0][0] = 0xC0;
		sbuf[0][1] = (uint8_t)(chCode >> 8);
		sbuf[0][2] = (uint8_t)(chCode & 0xFF);
		sbuf[0][3] = 0xB2;
		sbuf[0][4] = 0x08;
	
		// 2回目リクエスト設定
		sbuf[1][0] = 0xC0;
		sbuf[1][1] = (uint8_t)(chCode >> 8);
		sbuf[1][2] = (uint8_t)(chCode & 0xFF);
		sbuf[1][3] = 0x9A;
		sbuf[1][4] = 0x50;
	
		// 1回目のリクエスト送信
		IOUSBDevRequest req1;
		req1.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
		req1.bRequest = 0x03;
		req1.wValue = 0x3000;
		req1.wIndex = 0x00FE;
		req1.wLength = sizeof(sbuf[0]);
		req1.wLenDone = 0;
		req1.pData = sbuf[0];
		error = (*interface)->ControlRequest(interface, 0, &req1);
		if(error) {
			printf("setChannel: Unable to set the channel.\n");
			return 0;
		}
	
		// 2回目のリクエスト送信
		IOUSBDevRequest req2;
		req2.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
		req2.bRequest = 0x03;
		req2.wValue = 0x3000;
		req2.wIndex = 0x00FE;
		req2.wLength = sizeof(sbuf[1]);
		req2.wLenDone = 0;
		req2.pData = sbuf[1];
		error = (*interface)->ControlRequest(interface, 0, &req2);
		if(error) {
			printf("setChannel: Unable to set the channel.\n");
			return 0;
		}	
	
		// 3回目のリクエスト送信
		IOUSBDevRequest req3;
		req3.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
		req3.bRequest = 0x02;
		req3.wValue = 0x3000;
		req3.wIndex = 0x00B0;
		req3.wLength = sizeof(rbuf[0]);
		req3.wLenDone = 0;
		req3.pData = rbuf[0];
		error = (*interface)->ControlRequest(interface, 0, &req3);
		if(error) {
			printf("setChannel: Unable to set the channel.\n");
			return 0;
		}
	
		// 4回目のリクエスト送信
		IOUSBDevRequest req4;
		req4.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
		req4.bRequest = 0x02;
		req4.wValue = 0x3000;
		req4.wIndex = 0x0080;
		req4.wLength = sizeof(rbuf[1]);
		req4.wLenDone = 0;
		req4.pData = rbuf[1];
		error = (*interface)->ControlRequest(interface, 0, &req4);
		if(error) {
			printf("setChannel: Unable to set the channel.\n");
			return 0;
		}
		return 1;
	}
	//黒の場合
	int iCh=0;
	if ((channel >= 13) && (channel <= 23)) {
		// BS
		iCh = channel - 13;
		[self setChannel_Black:iCh Space:0];
	} else if ((channel >= 24) && (channel <= 35)){
		// 110CS
		iCh = channel - 24;
		[self setChannel_Black:iCh Space:1];
	} else {
		// 非対応
		printf("setChannel: Invalid channel.\n");
		return 0;
	}
	return 1;
}


// 黒専用チャンネル制御メソッド
// -----------------------------------------------------------
- (int)setChannel_Black:(int)channel Space:(uint32_t)dwSpace
{
	IOReturn error;


	struct Black_Channel_Sequence cmd[] = {
														// Index(n回目-1)
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x00, 0x0A },	// 0
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x01, 0x4C },	// 1
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x02, 0x02 },	// 2
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x03, 0x40 },	// 3
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x04, 0x6C },	// 4
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x05, 0x47 },	// 5
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x06, 0x02 },	// 6
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x07, 0x30 },	// 7
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x08, 0x20 },	// 8
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x09, 0xA2 },	// 9
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0A, 0xF1 },	// 10
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0B, 0x38 },	// 11
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0C, 0xD0 },	// 12
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0D, 0x10 },	// 13
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0E, 0xB0 },	// 14
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0F, 0x01 },	// 15
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x00, 0x0A },	// 16
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x01, 0x4C },	// 17
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x02, 0x02 },	// 18
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x03, 0x40 },	// 19
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x04, 0x6C },	// 20
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x05, 0x47 },	// 21
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x06, 0x02 },	// 22
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x07, 0x30 },	// 23
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x08, 0x20 },	// 24
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x09, 0xA2 },	// 25
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0A, 0xF1 },	// 26
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0B, 0x38 },	// 27
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0C, 0xD0 },	// 28
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0D, 0x10 },	// 29
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0E, 0xB0 },	// 30
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0F, 0x01 },	// 31
		{ 0x01, 0x3210, 0x0001, FALSE, 0x00, 0x00, 0x00 },	// 32
		{ 0x01, 0x3200, 0x0002, FALSE, 0x00, 0x00, 0x00 },	// 33
		{ 0x01, 0x3200, 0x0003, FALSE, 0x00, 0x00, 0x00 },	// 34
		{ 0x01, 0x3200, 0x0008, FALSE, 0x00, 0x00, 0x00 },	// 35
		{ 0x01, 0x3259, 0x000C, FALSE, 0x00, 0x00, 0x00 },	// 36
		{ 0x01, 0x32F2, 0x000D, FALSE, 0x00, 0x00, 0x00 },	// 37
		{ 0x01, 0x3251, 0x000F, FALSE, 0x00, 0x00, 0x00 },	// 38
		{ 0x01, 0x32B2, 0x0010, FALSE, 0x00, 0x00, 0x00 },	// 39
		{ 0x01, 0x3204, 0x001A, FALSE, 0x00, 0x00, 0x00 },	// 40
		{ 0x01, 0x3204, 0x001B, FALSE, 0x00, 0x00, 0x00 },	// 41
		{ 0x01, 0x3240, 0x0038, FALSE, 0x00, 0x00, 0x00 },	// 42
		{ 0x01, 0x3210, 0x0039, FALSE, 0x00, 0x00, 0x00 },	// 43
		{ 0x01, 0x3208, 0x003C, FALSE, 0x00, 0x00, 0x00 },	// 44
		{ 0x01, 0x3296, 0x003D, FALSE, 0x00, 0x00, 0x00 },	// 45
		{ 0x01, 0x32D9, 0x003E, FALSE, 0x00, 0x00, 0x00 },	// 46
		{ 0x01, 0x32B6, 0x003F, FALSE, 0x00, 0x00, 0x00 },	// 47
		{ 0x01, 0x325D, 0x0040, FALSE, 0x00, 0x00, 0x00 },	// 48
		{ 0x01, 0x3299, 0x0041, FALSE, 0x00, 0x00, 0x00 },	// 49
		{ 0x01, 0x3258, 0x0044, FALSE, 0x00, 0x00, 0x00 },	// 50
		{ 0x01, 0x3254, 0x0045, FALSE, 0x00, 0x00, 0x00 },	// 51
		{ 0x01, 0x3230, 0x0046, FALSE, 0x00, 0x00, 0x00 },	// 52
		{ 0x01, 0x3241, 0x0047, FALSE, 0x00, 0x00, 0x00 },	// 53
		{ 0x01, 0x3216, 0x0049, FALSE, 0x00, 0x00, 0x00 },	// 54
		{ 0x01, 0x3241, 0x004A, FALSE, 0x00, 0x00, 0x00 },	// 55
		{ 0x01, 0x3241, 0x004B, FALSE, 0x00, 0x00, 0x00 },	// 56
		{ 0x01, 0x325E, 0x004F, FALSE, 0x00, 0x00, 0x00 },	// 57
		{ 0x01, 0x3204, 0x0050, FALSE, 0x00, 0x00, 0x00 },	// 58
		{ 0x01, 0x32C0, 0x0051, FALSE, 0x00, 0x00, 0x00 },	// 59
		{ 0x01, 0x328A, 0x0052, FALSE, 0x00, 0x00, 0x00 },	// 60
		{ 0x01, 0x3213, 0x0053, FALSE, 0x00, 0x00, 0x00 },	// 61
		{ 0x01, 0x3290, 0x0054, FALSE, 0x00, 0x00, 0x00 },	// 62
		{ 0x01, 0x3292, 0x0055, FALSE, 0x00, 0x00, 0x00 },	// 63
		{ 0x01, 0x326D, 0x0058, FALSE, 0x00, 0x00, 0x00 },	// 64
		{ 0x01, 0x326F, 0x0059, FALSE, 0x00, 0x00, 0x00 },	// 65
		{ 0x01, 0x322E, 0x005A, FALSE, 0x00, 0x00, 0x00 },	// 66
		{ 0x01, 0x3223, 0x005B, FALSE, 0x00, 0x00, 0x00 },	// 67
		{ 0x01, 0x3233, 0x005D, FALSE, 0x00, 0x00, 0x00 },	// 68
		{ 0x01, 0x3204, 0x0087, FALSE, 0x00, 0x00, 0x00 },	// 69
		{ 0x01, 0x3200, 0x008D, FALSE, 0x00, 0x00, 0x00 },	// 70
		{ 0x01, 0x3000, 0x0077, FALSE, 0x00, 0x00, 0x00 },	// 71
		{ 0x01, 0x3000, 0x0071, FALSE, 0x00, 0x00, 0x00 },	// 72
		{ 0x01, 0x3000, 0x0071, FALSE, 0x00, 0x00, 0x00 },	// 73
		{ 0x01, 0x3240, 0x008F, FALSE, 0x00, 0x00, 0x00 },	// 74
		{ 0x01, 0x32F2, 0x0090, FALSE, 0x00, 0x00, 0x00 },	// 75
		{ 0x01, 0x3207, 0x00A5, FALSE, 0x00, 0x00, 0x00 },	// 76
		{ 0x01, 0x3207, 0x00A5, FALSE, 0x00, 0x00, 0x00 },	// 77
		{ 0x01, 0x32FF, 0x00A3, FALSE, 0x00, 0x00, 0x00 },	// 78
		{ 0x01, 0x32F7, 0x00A3, FALSE, 0x00, 0x00, 0x00 },	// 79
		{ 0x01, 0x3040, 0x0001, FALSE, 0x00, 0x00, 0x00 },	// 80
		{ 0x01, 0x3030, 0x0004, FALSE, 0x00, 0x00, 0x00 },	// 81
		{ 0x01, 0x3040, 0x0005, FALSE, 0x00, 0x00, 0x00 },	// 82
		{ 0x01, 0x3040, 0x0007, FALSE, 0x00, 0x00, 0x00 },	// 83
		{ 0x01, 0x30FF, 0x000F, FALSE, 0x00, 0x00, 0x00 },	// 84
		{ 0x01, 0x3027, 0x0011, FALSE, 0x00, 0x00, 0x00 },	// 85
		{ 0x01, 0x302F, 0x0013, FALSE, 0x00, 0x00, 0x00 },	// 86
		{ 0x01, 0x3031, 0x0014, FALSE, 0x00, 0x00, 0x00 },	// 87
		{ 0x01, 0x3002, 0x0016, FALSE, 0x00, 0x00, 0x00 },	// 88
		{ 0x01, 0x3020, 0x0022, FALSE, 0x00, 0x00, 0x00 },	// 89
		{ 0x01, 0x3070, 0x0024, FALSE, 0x00, 0x00, 0x00 },	// 90
		{ 0x01, 0x3079, 0x002C, FALSE, 0x00, 0x00, 0x00 },	// 91
		{ 0x01, 0x3034, 0x002D, FALSE, 0x00, 0x00, 0x00 },	// 92
		{ 0x01, 0x3028, 0x0030, FALSE, 0x00, 0x00, 0x00 },	// 93
		{ 0x01, 0x3031, 0x0031, FALSE, 0x00, 0x00, 0x00 },	// 94
		{ 0x01, 0x3001, 0x0038, FALSE, 0x00, 0x00, 0x00 },	// 95
		{ 0x01, 0x3078, 0x0039, FALSE, 0x00, 0x00, 0x00 },	// 96
		{ 0x01, 0x3090, 0x0048, FALSE, 0x00, 0x00, 0x00 },	// 97
		{ 0x01, 0x3068, 0x0051, FALSE, 0x00, 0x00, 0x00 },	// 98
		{ 0x01, 0x3038, 0x005E, FALSE, 0x00, 0x00, 0x00 },	// 99
		{ 0x01, 0x3021, 0x00C0, FALSE, 0x00, 0x00, 0x00 },	// 100
		{ 0x01, 0x3010, 0x00C1, FALSE, 0x00, 0x00, 0x00 },	// 101
		{ 0x01, 0x301A, 0x00E4, FALSE, 0x00, 0x00, 0x00 },	// 102
		{ 0x01, 0x301F, 0x00EA, FALSE, 0x00, 0x00, 0x00 },	// 103
		{ 0x00, 0x0000, 0x0000, FALSE, 0x00, 0x00, 0x00 },	// End
	};
	
	// チューニングシーケンス可変部分
	static const uint8_t PllConfA[] = {0x08, 0x08, 0x08, 0x08, 0x09, 0x09, 0x09, 0x0A, 0x0A, 0x0A, 0x0A, 0x0C, 0x0C, 0x0D, 0x0D, 0x0D, 0x0E, 0x0E, 0x0E, 0x0F, 0x0F, 0x0F, 0x10};	// 1/17回目 +40
	static const uint8_t PllConfB[] = {0x32, 0x32, 0x80, 0x80, 0x64, 0x64, 0x64, 0x00, 0x00, 0x4C, 0x4C, 0x9A, 0xEA, 0x3A, 0x8A, 0xDA, 0x2A, 0x7A, 0xCA, 0x1A, 0x6A, 0xBA, 0x0A};	// 2/18回目 +40
	static const uint8_t PllConfC[] = {0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};	// 75回目	+02
	static const uint8_t PllConfD[] = {0x10, 0x11, 0x30, 0x31, 0x90, 0x91, 0x92, 0xD0, 0xD1, 0xF1, 0xF2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};	// 76回目	+02
	
	const uint32_t dwChIndex = dwSpace * 11 + channel;
	
	// 資料
	// BSAT-2c/3a TsID=4[BSch番号(2byte)][TP番号]
	// BS- 1  周波数=11.727GHz 0832 TsID=4010 BS朝日
	// BS- 1  周波数=11.727GHz 0832 TsID=4011 BS-i
	// BS- 3  周波数=11.765GHz 0880 TsID=4030 WOWWOW
	// BS- 3  周波数=11.765GHz 0880 TsID=4031 BSジャパン
	// BS- 5                                   アナログで使用(WOWWOW)
	// BS- 7                                   アナログで使用(NHK BS-1)
	// BS- 9  周波数=11.880GHz 0964 TsID=4090 BS11
	// BS- 9  周波数=11.880GHz 0964 TsID=4091 スターチャンネル
	// BS- 9  周波数=11.880GHz 0964 TsID=4092 TwellV
	// BS-11                                   アナログで使用(NHK BS-2)
	// BS-13  周波数=11.957GHz 0A00 TsID=40D0 BS日テレ
	// BS-13  周波数=11.957GHz 0A00 TsID=40D1 BSフジ
	// BS-15  周波数=11.996GHz 0A4C TsID=40F1 NHK BS1/BS2/ウェザーニューズ/エンジニアリングストリーム
	// BS-15  周波数=11.996GHz 0A4C TsID=40F2 NHK BS-Hi
	// BS-17                                   未使用(BSAT-4以降に搭載予定)
	// BS-19                                   未使用(BSAT-4以降に搭載予定)
	// BS-21                                   未使用(BSAT-4以降に搭載予定)
	// BS-23                                   未使用(BSAT-4以降に搭載予定)
	//
	// NSAT-110 TsID=0000 固定
	// CSND2  周波数=12.291GHz 0C9A TsID=0000 CS
	// CSND4  周波数=12.331GHz 0CEA TsID=0000 CS
	// CSND6  周波数=12.371GHz 0D3A TsID=0000 CS
	// CSND8  周波数=12.411GHz 0D8A TsID=0000 CS
	// CSND10 周波数=12.451GHz 0DDA TsID=0000 CS
	// CSND12 周波数=12.491GHz 0E2A TsID=0000 CS
	// CSND14 周波数=12.531GHz 0E7A TsID=0000 CS
	// CSND16 周波数=12.571GHz 0ECA TsID=0000 CS
	// CSND18 周波数=12.611GHz 0F1A TsID=0000 CS
	// CSND20 周波数=12.651GHz 0F6A TsID=0000 CS
	// CSND22 周波数=12.691GHz 0FBA TsID=0000 CS
	// CSND24 周波数=12.731GHz 100A TsID=0000 CS
	
	uint32_t dwXmitNum = 0;
	for (;;dwXmitNum++) {
		struct Black_Channel_Sequence tmp = cmd[dwXmitNum];
		if (tmp.byReqNo == 0x00) break; // 終端
		
		// リクエストの作成
		uint8_t  buf[3]  = "";
		uint32_t buf_len = 0;
		
		if (tmp.bData) {
			buf_len = 3;
			buf[0]  = tmp.byData1;
			buf[1]  = tmp.byData2;
			buf[2]  = tmp.byData3;
		}

		IOUSBDevRequest ctrl;
		ctrl.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
		ctrl.bRequest = tmp.byReqNo;
		ctrl.wValue = tmp.wValue;
		ctrl.wIndex = tmp.wIndex;
		ctrl.wLength = buf_len;
		ctrl.wLenDone = 0;
		ctrl.pData = buf;
		
		// 可変部の変更
		switch(dwXmitNum){
			case 0:		buf[2] = PllConfA[dwChIndex];	break;	// 周波数1 データの最終byteにいれる
			case 1:		buf[2] = PllConfB[dwChIndex];	break;	// 周波数2 データの最終byteにいれる
			case 16:	buf[2] = PllConfA[dwChIndex];	break;	// 周波数1 データの最終byteにいれる
			case 17:	buf[2] = PllConfB[dwChIndex];	break;	// 周波数2 データの最終byteにいれる
			case 74:	ctrl.wValue = (ctrl.wValue & 0xFF00) | ((uint16_t)PllConfC[dwChIndex]);		break;	// TsID1 wValueの下位Byteに入れる
			case 75:	ctrl.wValue = (ctrl.wValue & 0xFF00) | ((uint16_t)PllConfD[dwChIndex]);		break;	// TsID2 wValueの下位Byteに入れる
		}
		
		// 送信
		error = (*interface)->ControlRequest(interface, 0, &ctrl);
		if(error) {
			printf("setChannel: Unable to set the channel.\n");
			return 0;
		}		
		
		// 何故かここで遅延をかけないとTimeout errorが起こる。
		//usleep(SETCHANNEL_COMMAND_INTERBAL * 1000);
	}

	return 1;
}



// シグナルレベル取得メソッド
// -----------------------------------------------------------
- (float)getSignalLevel
{
	IOReturn err;
	if( friio_type == FRIIO_WHITE ){
		//白用
		float signal_level = 0;
		uint8_t sigbuf[37];
		uint32_t dwHex;
		
		// リクエストバッファ
		memset(sigbuf, '\0', sizeof(sigbuf));
	
		// リクエスト送信
		IOUSBDevRequest req;
		req.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
		req.bRequest = 0x02;
		req.wValue = 0x3000;
		req.wIndex = 0x0089;
		req.wLength = sizeof(sigbuf);
		req.wLenDone = 0;
		req.pData = sigbuf;
		err = (*interface)->ControlRequest(interface, 0, &req);
		if(err) {
			printf("setChannel: Unable to get the signal level.\n");
			return 0;
		}	

		// シグナルレベルを算出
		dwHex = ((uint32_t)(sigbuf[2]) << 16) | ((uint32_t)(sigbuf[3]) << 8) | (uint32_t)(sigbuf[4]);
		if(!dwHex) {
			signal_level = 0.0f;
		} else {
			signal_level = (float)(68.4344993f - log10((double)dwHex) * 10.0 + pow(log10((double)dwHex) - 5.045f, (double)4.0f) / 4.09f);
		}
	
		return signal_level;	
	}
	//黒用
	// 線形補完で近似する
	static const float afLevelTable[] =
	{
		24.07f,	// 00	00	0		24.07dB
		24.07f,	// 10	00	4096	24.07dB
		18.61f,	// 20	00	8192	18.61dB
		15.21f,	// 30	00	12288	15.21dB
		12.50f,	// 40	00	16384	12.50dB
		10.19f,	// 50	00	20480	10.19dB
		8.140f,	// 60	00	24576	8.140dB
		6.270f,	// 70	00	28672	6.270dB
		4.550f,	// 80	00	32768	4.550dB
		3.730f,	// 88	00	34816	3.730dB
		3.630f,	// 88	FF	35071	3.630dB
		2.940f,	// 90	00	36864	2.940dB
		1.420f,	// A0	00	40960	1.420dB
		0.000f	// B0	00	45056	-0.01dB
	};
	
	uint8_t sigbuf[4];

	// リクエストバッファ
	memset(sigbuf, '\0', sizeof(sigbuf));

	// リクエスト送信
	IOUSBDevRequest req;
	req.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
	req.bRequest = 0x02;
	req.wValue = 0x3200;
	req.wIndex = 0x00BA;
	req.wLength = 4;//sizeof(sigbuf);
	req.wLenDone = 0;
	req.pData = sigbuf;
	err = (*interface)->ControlRequest(interface, 0, &req);
	if(err) {
		printf("setChannel: Unable to get the signal level.\n");
		return 0;
	}

	// 信号レベル計算
	if(sigbuf[2] <= 0x10){
		// 最大クリップ
		return 24.07f;
	} else if (sigbuf[2] >= 0xB0) {
		// 最小クリップ
		return 0.0f;
	}
	// 線形補完
	const float fMixRate = (float)(((uint16_t)(sigbuf[2] & 0x0F) << 8) | (uint16_t)sigbuf[3]) / 4095.0f;
	return afLevelTable[sigbuf[2] >> 4] * (1.0f - fMixRate) + afLevelTable[(sigbuf[2] >> 4) + 0x01U] * fMixRate;
}


// LED制御メソッド
// -----------------------------------------------------------
- (void)setLED:(int)color
{
	[self _sendRequest:0x000F withIndex:0x06];
	switch(color) {
		// 紫
		case 0:
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0009 withIndex:0x00];
			[self _sendRequest:0x000D withIndex:0x00];
			break;
		// 緑
		case 1:
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0001 withIndex:0x00];
			[self _sendRequest:0x0005 withIndex:0x00];
			break;
		// 赤
		case 2:
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x000B withIndex:0x00];
			[self _sendRequest:0x000F withIndex:0x00];
			[self _sendRequest:0x0003 withIndex:0x00];
			[self _sendRequest:0x0007 withIndex:0x00];
			[self _sendRequest:0x0001 withIndex:0x00];
			[self _sendRequest:0x0005 withIndex:0x00];
			break;
		// 白
		case 3:
			// 情報なし
			break;
	}
	[self _sendRequest:0x0000 withIndex:0x06];
}


@end
